/*! ********************************************
 * SynthData.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthData_h
#define INCLUDE_SynthData_h

#ifndef INCLUDE_GrayBox_Types_h
#include "Types.h"
#endif

namespace sora {
namespace graybox {
namespace synth {
class SynthData final
{
// Instance ---
private:
	static SynthData* _instance;
public:
	static SynthData* getInstance();
	static void disposeInstance();
private:
// Constructor
	SynthData() = default;
// Copy Constructor
	SynthData(const SynthData&) = delete;
	SynthData& operator = (const SynthData&) = delete;
// Move Constructor
	SynthData(SynthData&&) = delete;
	SynthData& operator = (SynthData&&) = delete;
// Destructor
	~SynthData();
// Const ---
	// Volume, Pan
	// Log2 Volume Table
public:
	static constexpr int32 LOG2VT_DYN_BITS     = 16;                    // const 16    96dB dynamic range (FM TL, 16Bit PCM)
	static constexpr int32 LOG2VT_DYN          = 1 << LOG2VT_DYN_BITS;  //             1bit per 6dBv
	static constexpr int32 LOG2VT_DYN_RES_BITS = 8;
	static constexpr int32 LOG2VT_DYN_RES      = 1 << LOG2VT_DYN_RES_BITS;
	static constexpr int32 LOG2VT_SIZE         = LOG2VT_DYN_BITS << (LOG2VT_DYN_RES_BITS + 1);
	// Pow2 Table
	static constexpr int32 POW2T_OCTV_BITS = 13;  // const 13
	static constexpr int32 POW2T_OCTV_RES  = 1 << POW2T_OCTV_BITS;  // POW2T_OCTV_RES == LOG2VT_SIZE
	static constexpr int32 POW2T_OCTV_MASK = POW2T_OCTV_RES - 1;
	// Envelope Generator
	static constexpr int32 EG_MAX_LEVEL_BITS = 4 + LOG2VT_DYN_RES_BITS;  // constant 12 (4 = 16 LOG2VT_DYN_BITS)
	static constexpr int32 EG_MAX_LEVEL      = 1 << EG_MAX_LEVEL_BITS;   // 4096 = 16 * 256
	static constexpr int32 EG_MAX_CNT_BITS   = 30;
	static constexpr int32 EG_MAX_CNT        = 1 << EG_MAX_CNT_BITS;
	static constexpr int32 EG_CNT_FP         = EG_MAX_CNT_BITS - EG_MAX_LEVEL_BITS;  // for EGLevelTable.  xbit fixed-point  (precision 1 / (1 << x))
	static constexpr int32 EG_ENV_DYN_BITS = 13;  // 1 ~ 16
	static constexpr int32 EG_ENV_DYN      = 1 << EG_ENV_DYN_BITS;
	static constexpr int32 EG_CNT_ENV_ATK  = int(1.0 / EG_ENV_DYN * EG_MAX_LEVEL) << EG_CNT_FP;
	static constexpr int32 EG_CNT_ENV_BTM  = (EG_ENV_DYN_BITS * LOG2VT_DYN_RES) << EG_CNT_FP;
	static constexpr int32 EG_SUSTAIN_BTM  = EG_ENV_DYN_BITS * 8;
	static constexpr double EG_REFERENCE_TIME = 16.0;                           // to 1/2 Volt (-6dB) Time [EG : Rate  0]
	static constexpr double EG_SOUND_OFF_TIME = 1.0 / (EG_REFERENCE_TIME * 2.0);  // to 1/2 Volt (-6dB) Time [EG : Rate 34]
	// LFO
	static constexpr double LFO_MAX_FREQ        = 32.0;  // Hz
	static constexpr int32 LFO_TIMER_INIT_VALUE = 0x40000000;  // initial Value - Phase Update Timer
	static constexpr int32 LFO_WAVE_TABLE_SIZE_BITS = 8;
	static constexpr int32 LFO_WAVE_TABLE_SIZE      = 1 << LFO_WAVE_TABLE_SIZE_BITS;  // PM[0,2,4...],AM[1,3,5...]
	// Phase Modulation
	
// Tables ---
	// Volume, Pan
private:
	float* _VOLUME_TABLE_0dB = nullptr;
	float* _VOLUME_TABLE_6dB = nullptr;
	float* _PAN_TABLE_SIN    = nullptr;
	float* _PAN_TABLE_BAL    = nullptr;
	// Log2 Volume Table
	float* _LOG2V_TABLE   = nullptr;
	int32* _LOG2V_TABLE_I = nullptr;
	// Pow2 Table
	int32* _POW2_TABLE      = nullptr;
	int32* _POW2T_RS_TABLE  = nullptr;
	int32* _POW2T_KEY_TABLE = nullptr;
	// Envelope Generator
	int32* _EG_TOTAL_LEVEL_TABLE = nullptr;
	int32* _EG_LEVEL_TABLE_UP    = nullptr;
	int32* _EG_LEVEL_TABLE_DW    = nullptr;
	int32* _EG_KEY_TABLE      = nullptr;
	int32* _EG_POW2T_RS_TABLE = nullptr;
	int32* _EG_CNT_GATE_TABLE = nullptr;
	// LFO
	int32* _LFO_WAVE_TABLES = nullptr;
	// Phase Modulation
	int32 * _MUL_TABLE = nullptr;
	
// getter, creator ---
// Volume, Pan ==================================================
private:
	float* _createPAN_TABLE(int32 type);
public:
	static float* createVOLUME_TABLE(int32 size, float c = 0.0f, bool reverse = false);
	float* getVOLUME_TABLE_0dB();
	float* getVOLUME_TABLE_6dB();
	float* getPAN_TABLE_SIN();
	float* getPAN_TABLE_BAL();
// Log2 Volume Table ==================================================
private:
	static constexpr float  _LOG2VT_DYN_RESf = float(-LOG2VT_DYN_RES);
	static constexpr double _LOG2VT_DYN_RESd = double(-LOG2VT_DYN_RES);
	static constexpr float  _LOG2VT_BOADERf  = 1.0f / (1 << LOG2VT_DYN_BITS);
	static constexpr double _LOG2VT_BOADERd  = 1.0 / (1 << LOG2VT_DYN_BITS);
public:
	static int32 toLog2vtIndexValue(float vol);
	static int32 toLog2vtIndexValue(double vol);
	float* getLOG2V_TABLE();
	int32* getLOG2V_TABLE_I();
// Pow2 Table ==================================================
private:
	static constexpr float  _POW2T_OCTV_RESf = float(POW2T_OCTV_RES);
	static constexpr double _POW2T_OCTV_RESd = double(POW2T_OCTV_RES);
public:
	static int32 toPow2tIndexValue(float val);
	static int32 toPow2tIndexValue(double val);
	int32* getPOW2_TABLE();
	int32* getPOW2T_RS_TABLE();
	int32* getPOW2T_KEY_TABLE();
// Envelope Generator ==================================================
public:
	int32* getEG_TOTAL_LEVEL_TABLE();
	int32* getEG_LEVEL_TABLE_DW();
	int32* getEG_LEVEL_TABLE_UP();
	int32* getEG_KEY_TABLE();
	int32* getEG_POW2T_RS_TABLE();
	int32* getEG_CNT_GATE_TABLE();
// LFO ==================================================
private:
	static constexpr float LFO_MAX_FREQf = float(LFO_MAX_FREQ);  // Hz
public:
	static int32 lfoTimerCoef(float sampleRate, int32 steps);
	static int32 lfoTimerCoef(double sampleRate, int32 steps);
	int32* getLFO_WAVE_TABLES();
// Phase Modulation ==================================================
public:
	int32* getMUL_TABLE();

};
}}}
#endif